/*
 * $Id: kom-types.h,v 0.14 1994/12/31 11:43:29 ceder Exp $
 * Copyright (C) 1990, 1991, 1992, 1993, 1994  Lysator Academic Computer Association.
 *
 * This file is part of the LysKOM server.
 * 
 * LysKOM is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 1, or (at your option) 
 * any later version.
 * 
 * LysKOM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with LysKOM; see the file COPYING.  If not, write to
 * Lysator, c/o ISY, Linkoping University, S-581 83 Linkoping, SWEDEN,
 * or the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, 
 * MA 02139, USA.
 *
 * Please mail bug reports to bug-lyskom@lysator.liu.se. 
 */
/*
 *  KOM-types.h  --  Types used by both server and client of LysKOM
 *
 *
 *  Copyright (C) 1990, 1991, 1992, 1993, 1994	Lysator Computer Club,
 *			Linkoping University,  Sweden
 *
 *  Everyone is granted permission to copy, modify and redistribute
 *  this code, provided the people they give it to can.
 *
 *
 *  Filecreator: Thomas Bellman
 *		  Lysator Computer Club
 *		  Linkoping University
 *		  Sweden
 *
 *  email:	  Bellman@Lysator.LiU.SE
 */



#ifndef  KOM_TYPES_H_ALREADY_INCLUDED__

#define  KOM_TYPES_H_ALREADY_INCLUDED__

#ifndef SERVER
#  error must define SERVER
#endif

#ifdef CLIENT
#  error This file no longer supports CLIENT
#endif

typedef unsigned short Pers_no;
typedef unsigned short Conf_no;
typedef unsigned long  Text_no;
typedef unsigned long  Local_text_no;
typedef unsigned long  Session_no;

#ifdef PROT_a

/*
 * To uniquely identify a person, conference or text, three numbers
 * are used: Universe, Site and X_{recpt,person,text}. The universe is
 * omitted when it is LysKOM, or is implied by a echo conference.
 */
/*
 * LysKOM, RFC-822, FidoNet, News are examples of universes.
 */
/*typedef		unsigned long		Universe;*/

/*
 * JAKOM, lysator.liu.se, 2:504/113, foo.bar.com are examples of sites
 * in the above universes. (Note that comp.lang.c is _not_ a site in
 * the News universe).
 */
/*typedef		unsigned long		Site;*/

/*
 * Fritt Forum, ceder, Per Cederqvist, comp.lang.c are examples of
 * x_recpt in the above universes.
 */

/*
 * But for now, we have a System_id (sum of Universe and Site) and a
 * Global_x.
 */

typedef		unsigned long		System_id;
typedef		unsigned long		Global_recpt;
typedef		unsigned long		Global_person;
typedef		unsigned long		Global_text;

#endif


#define	MAX_PERS_NO		((Pers_no) USHRT_MAX)
#define MAX_CONF_NO		((Conf_no) USHRT_MAX)
#define MAX_TEXT_NO		((Text_no) ULONG_MAX)
#define	MAX_LOCAL_TEXT_NO	((Local_text_no) ULONG_MAX)


#define PASSWD_LEN 64
typedef char   Password[PASSWD_LEN];
typedef off_t  Text_index;
typedef time_t Time;


/*
 * The privilige bits:  says what priviliges a person has.  E g
 *  if he is allowed to read texts he normally shouldn't be allowed
 *  to read. See file doc/security-levels.txt
 */
typedef	struct {
	unsigned int	wheel	: 1;
	unsigned int	admin	: 1;
	unsigned int	statistic : 1;
	unsigned int	create_pers: 1;
	unsigned int	create_conf: 1;
	unsigned int	change_name: 1;
	unsigned int	extern_gw : 1;
	unsigned int	flg8	: 1;	/* For future use. */
	unsigned int	flg9    : 1;
	unsigned int	flg10   : 1;
	unsigned int	flg11   : 1;
	unsigned int	flg12	: 1;
	unsigned int	flg13	: 1;
	unsigned int	flg14	: 1;
	unsigned int	flg15	: 1;
	unsigned int	flg16	: 1;
} Priv_bits;

/*
 * Flags in the Person struct.
 */
typedef struct {
    unsigned int  unread_is_secret : 1;/* FALSE if everyone is allowed to ask
					  how many unread texts you have. */
    unsigned int  flg2	: 1;
    unsigned int  flg3	: 1;
    unsigned int  flg4	: 1;
    unsigned int  flg5	: 1;
    unsigned int  flg6	: 1;
    unsigned int  flg7	: 1;
    unsigned int  flg8	: 1;
} Personal_flags;


/* See file doc/misc_items */
typedef enum {
    recpt     = 0,	/* 0 Recipient (conference) */
    cc_recpt  = 1,	/* 1 Carbon Copy recipient (extra kopia) */
    comm_to   = 2,	/* 2 Comment to (text) */
    comm_in   = 3,	/* 3 Commented in (text) */
    footn_to  = 4,	/* 4 This is a footnote to (text) */
    footn_in  = 5,	/* 5 Footnote to this text in (text) */
#ifdef PROT_a
    x_author  = 10,	/* 10 External author (Global_name). */
#endif
    
    loc_no    = 6,	/* 6 Sequence number within conference */
    rec_time  = 7,	/* 7 Received at (time) */
    sent_by   = 8,	/* 8 Sent by (person) */
    sent_at   = 9	/* 9 Sent at (time) */
#ifdef PROT_a
	,
    x_person  = 11,	/* 11 External person id. (Global_name). */
    x_recpt   = 12,	/* 12 External recipient. (Global_name). */
    x_text    = 13,	/* 13 External text. (Global_text). */
    x_system  = 14	/* 14 External system identification (System_id). */
#endif
} Info_type;

#ifdef PROT_a
typedef String Global_name;
#endif

typedef	union {
		Conf_no		recipient;
		Conf_no		cc_recipient;
		Text_no		comment_to;
		Text_no		commented_in;
		Text_no		footnote_to;
		Text_no		footnoted_in;

#ifdef PROT_a
		Global_name	x_author; /* Protocol a. */
#endif

		Local_text_no	local_no;
		Time		received_at;
		Pers_no		sender;
		Time		sent_at;

#ifdef PROT_a
		Global_name	global_person;
		Global_name	global_recpt;
		Global_text	global_text;
		System_id	system_id;
#endif
} Info_datum;


/*  This struct contains miscellaneous information about a text.  */
typedef	struct {
	Info_type	type;
	Info_datum	datum;
} Misc_info;


/*  Fields of this type is supposed to tell the garbage collector
 *  which texts it should remove first.
 */
typedef unsigned long	Garb_nice;


/*  Struct for text status  */
typedef struct {
	Time		  creation_time;
#ifdef SERVER
        long              file_pos; /* Start of the text in the text file. */
#endif
	Pers_no		  author;
	unsigned short		  no_of_lines;
	String_size	  no_of_chars;
	unsigned short		  no_of_marks;	/* Antal markeringar */
	unsigned short		  no_of_misc;	/* Recipients, times, comments, ... */
	Misc_info	* misc_items;	/* List of miscellaneous info. */
} Text_stat;


typedef	struct {
	unsigned int	rd_prot	: 1;	/* !(Can anyone become a member?) */
	unsigned int	original : 1;	/* Comments forbidden? */
	unsigned int	secret : 1;	/* Secret conference? */
				/* Note: if a letterbox is secret it
				 * will be very hard for that person
				 * to log in, since he can't map his
				 * name to a pers_no. He must either
				 * know his pers_no, or have another
				 * identity which is his supervisor.
				 */
	unsigned int	letter_box : 1;	/* Is this a letter box? */
} Conf_type;


typedef struct {
	Pers_no		member;
} Member;


/* Struct for marks */
typedef struct {
	Text_no		text_no;
	unsigned char	mark_type;	/* It's up to the clients to decide the
					   meaning of this field. */
} Mark;



/*  Information about a person's membership in a conference  */
typedef struct {
	Conf_no		conf_no;
	unsigned char	priority;	/* Interrupt priority */
	Time		last_time_read;	/* Updated every time a text in this
					   conf. is marked as read.	*/
	Local_text_no	last_text_read;	/* All texts before and inclusive this
					   are read	*/
	unsigned short	no_of_read;
	Local_text_no * read_texts;	/* Texts after last_text_read. Sorted
					   in ascending order */
} Membership;


/* Some structs to handle variable-sized arrays. */

typedef struct {
	unsigned short	  no_of_marks;
	Mark	* marks;		/* Pointer to an array of marks. */
} Mark_list;


typedef struct {
    unsigned short no_of_confs;
    Conf_no  *conf_nos;
} Conf_no_list;


typedef struct {
	unsigned short	  no_of_members;
	Member	* members;
} Member_list;


typedef  struct {
	Local_text_no	  first_local_no;
    	unsigned long		  no_of_texts;
	Text_no		* texts;
}  Text_list;


/*  A list of conference numbers, also telling if it is a
 *  mailbox (i e a person) or an ordinary conference.	*/
/* Delete this ugly variant as soon as possible! */
typedef struct {
	unsigned long		  no_of_conf_nos;
	Conf_no		* conf_nos;
	Conf_type	* type_of_conf;
} Conf_list_old;
/* This is the way it should look! */
typedef struct {
	Conf_no   conf_no;
	Conf_type type;
} Micro_conf;

typedef struct {
	unsigned long       no_of_confs;
	Micro_conf * confs;
} Conf_list;


/*  A list of person numbers  */
typedef struct {
	unsigned long		  no_of_persons;
	Pers_no		* persons;
} Pers_list;


typedef struct {
	unsigned short	      no_of_confs;
	Membership  * confs;
} Membership_list;


/* The Info struct */
typedef struct {
	long		version;
	Conf_no		conf_pres_conf; /* Presentation of new confs */
	Conf_no		pers_pres_conf;	/* Presentation of new persons */
	Conf_no		motd_conf; 	/* Conf that receive motds */
	Conf_no		kom_news_conf;  /* News about kom */
	Text_no		motd_of_lyskom; /* To be displayed after login */
	/* and maybe more... */
} Info;

/* For performance reasons, sort the fields in increasing size.
  The comments about number of bits are true on a Sun Sparc-2, not
  necessarily anything else.  The code should work anyhow. */
typedef	struct {
        /* 8-bit quantities */
	Conf_type	type;		/* secret, rd_prot etc */

	/* 16-bit quantities */
	Pers_no		creator;
	Conf_no		supervisor;	/* Organisat|r f|r m|tet */
	Conf_no		permitted_submitters;	/* People who are allowed
					   to submit texts to this conf. 
					   0 -> anyone may submit. */
	Conf_no		super_conf;	/* Send unallowed submissions to
					   the super_conf. 
					   0 -> unallowed submissions bounce */

	/* 32-bit quantities */
	Time		creation_time;
	Text_no		presentation;
	Time		last_written;	/* Time of latest text in this conf. */
	Text_no		msg_of_day;	/* Message to be displayed when this
					   conf is referenced by the user. */
	Garb_nice	nice;		/* How long do texts in this
					   conf live? */
	String		name;		/* Name of conference */
#ifdef CLIENT
	unsigned short		no_of_members;
	Local_text_no	first_local_no;
	unsigned long		no_of_texts;
#else
	Member_list	members;	/* List of members in conf */
	Text_list	texts;		/* List of texts */
#endif
} Conference;



typedef struct {
	String		name;		/* Name of conference */
	Conf_type	type;		/* secret, rd_prot etc */
	Local_text_no	highest_local_no; /* highest local text no */
	Garb_nice	nice;	        /* Number of days to live */
} Small_conf;

typedef struct {
	String      name;
	Conf_type   type;
	Conf_no     conf_no;
} Conf_z_info;

typedef struct {
	unsigned short    no_of_confs;
	Conf_z_info     * confs;
} Conf_z_info_list;

/*  Struct for persons  */
typedef struct {
	Text_no		user_area;	/* Misc info the clients might want to
					   store. 0 = not used. */
	Priv_bits	privileges;
	Personal_flags	flags;
	Time		last_login; 	/* Or logout */
	unsigned long		total_time_present; /* Number of seconds. */
	unsigned long		sessions;	/* Number of sessions */
	unsigned long		created_lines;	/* No. of created lines (statistics) */
	unsigned long		created_bytes;	/* No. of created bytes (statistics) */
	unsigned long		read_texts;	/* No. of read texts (statistics) */
	unsigned long		no_of_text_fetches; /* (statistics) */
	unsigned short		created_persons; /* (statistics) */
	unsigned short		created_confs;	/* (statistics) */
	String		username;	/* User-name & hostname */
#ifdef CLIENT
	Local_text_no	first_created_text; /* The first text that still
					     * exists. */
	unsigned long		no_of_created_texts;
	unsigned short		no_of_marks;	/* This many marked texts */
	unsigned short		no_of_confs;	/* Member in this many confs */
#else	/* SERVER */
	Text_list	created_texts;
	Mark_list	marks;		/* List of marked texts */
	Membership_list conferences;	/* List of conferences the person is
					 * a member in. */
	Password	pwd;		/* Encrypted password */
#endif
} Person;


/*
 * This struct is returned from the 'who_is_on' call.
 */
typedef struct {
	Pers_no		person;
	String		what_am_i_doing;
	Conf_no		working_conference;
} Who_info_old;

typedef struct {
	Pers_no		person;
	String		what_am_i_doing;
	String		username;           /* Userid and hostname. */
	Conf_no		working_conference;
	Session_no	session_no;         /* Serial number of connection. */
} Who_info;

typedef struct {
	Pers_no		person;
	String		what_am_i_doing;
	String		username;           /* Userid and hostname. */
	String		ident_user;	     /* According to Ident protocol. */
	String		hostname; 	     /* According to TCP/IP. */
	Conf_no		working_conference;
	Session_no	session_no;         /* Serial number of connection. */
} Who_info_ident;

typedef struct {
	Pers_no		person;
	String		what_am_i_doing;
	String		username;            /* Userid and hostname. */
	Conf_no		working_conference;
	Session_no	session;             /* Serial number of connection. */
	Time		connection_time;     /* Not logintime. */
	unsigned long		idle_time; 	     /* Seconds. */
} Session_info;


typedef struct {
	Pers_no		person;
	String		what_am_i_doing;
	String		username;            /* Userid and hostname,
						according to the client. */
	String		ident_user;	     /* According to Ident protocol. */
	String		hostname; 	     /* According to TCP/IP. */
	Conf_no		working_conference;
	Session_no	session;             /* Serial number of connection. */
	Time		connection_time;     /* Not logintime. */
	unsigned long		idle_time; 	     /* Seconds. */
} Session_info_ident;


typedef struct {
	int		  no_of_persons;
    	Who_info_old	* info;
} Who_info_list_old;	


typedef struct {
	int		  no_of_persons;
    	Who_info	* info;
} Who_info_list;	


typedef struct {
	int		  no_of_persons;
    	Who_info_ident	* info;
} Who_info_ident_list;

typedef struct {
        unsigned long protocol_version;
        String        server_name;
        String        server_version;
} Version_info;

#ifdef SERVER
typedef struct {
        unsigned long protocol_version;
        char *        server_name;
        char *        server_version;
} Version_info_internal;
#endif

#endif	/*  ifndef  KOM_TYPES_H_ALREADY_INCLUDED__  */
