/*
 * $Id: admin.c,v 0.18 1995/01/01 19:31:42 ceder Exp $
 * Copyright (C) 1991, 1993, 1994, 1995  Lysator Academic Computer Association.
 *
 * This file is part of the LysKOM server.
 * 
 * LysKOM is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 1, or (at your option) 
 * any later version.
 * 
 * LysKOM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with LysKOM; see the file COPYING.  If not, write to
 * Lysator, c/o ISY, Linkoping University, S-581 83 Linkoping, SWEDEN,
 * or the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, 
 * MA 02139, USA.
 *
 * Please mail bug reports to bug-lyskom@lysator.liu.se. 
 */
/*
 * admin.c
 *
 * Administrative calls.
 */

static char *rcsid = "$Id: admin.c,v 0.18 1995/01/01 19:31:42 ceder Exp $";
#include "rcs.h"
USE(rcsid);

#include <stdio.h>
#include <setjmp.h>
#ifdef HAVE_STDARG_H
#  include <stdarg.h>
#endif
#include <sys/types.h>
#include <time.h>

#include "misc-types.h"
#include "s-string.h"
#include "kom-types.h"
#include "manipulate.h"
#include "com.h"
#include "connections.h"
#include "kom-errno.h"
#include "cache.h"
#include "config.h"
#include "log.h"
#include "send-async.h"
#include "param.h"
#include "string-malloc.h"
#include "services.h"
#include "version-info.h"

/* All of the fields in this structure except the version number
   is set from the configuration file at startup (see ramkomd.c).
   The default values are also set in ramkomd.c; the values below are
   never used.  The defaults are as of this writing (1994-01-11
   19:57:22) equal to the values below, but don't trust that that is
   so when you read this. */
Info kom_info = 
{
#include "version.incl"
    ,				/* version */
    1,				/* conf_pres_conf */
    2,				/* pers_pres_conf */
    3,				/* motd_conf */
    4,				/* kom_news_conf */
    0				/* motd_of_lyskom */
};

/*
 * Return info about this server. This can (and should) be done
 * before logging in. motd_of_lyskom should be displayed before
 * prompting for username if it isn't 0.
 */
extern Success
get_info( Info *result )
{
    *result = kom_info;
    return OK;
}

extern Success
get_version_info( Version_info *result )
{
    /* Allowed before login. */
    result->protocol_version = kom_version_info.protocol_version;
    result->server_name = s_fcrea_str(kom_version_info.server_name);
    result->server_version = s_fcrea_str(kom_version_info.server_version);
    return OK;
}

/* /// */
extern Success
set_motd_of_lyskom (Text_no motd)
{
    Text_stat *old_motd = NULL;
    Text_stat *new_motd = NULL;

    CHK_LOGIN(FAILURE);

    if ( !ENA(admin, 1) )
    {
	kom_errno = KOM_PERM;
	return FAILURE;
    }
    
    /* Check that the new motd exists before deleting the old*/

    if ( motd != 0 )
    {
	GET_T_STAT(new_motd, motd, FAILURE);
	if ( new_motd->no_of_marks >= param.max_marks_text )
	{
	    log("LIMIT: set_motd_of_lyskom(%d): New motd has %d marks.\n",
		motd, new_motd->no_of_marks);
	    kom_errno = KOM_MARK_LIMIT;
	    return FAILURE;
	}
    }
    
    /* Unmark the previous motd if it exists. */

    if ( kom_info.motd_of_lyskom != 0
	&& (old_motd = cached_get_text_stat(kom_info.motd_of_lyskom)) != NULL)
    {
	if ( old_motd->no_of_marks > 0 )
	{
	    --old_motd->no_of_marks;
	    mark_text_as_changed( kom_info.motd_of_lyskom );
	}
	else
	{
	    log("ERROR: do_set_motd(): Old motd not marked\n");
	}
    }

    /* Mark the new motd */

    if ( motd != 0 )
    {
	++new_motd->no_of_marks;
	mark_text_as_changed( motd );
    }
    
    kom_info.motd_of_lyskom = motd;

    return OK;
}




/*
 * Force all clients to read a message.
 * Sends an asynchronous message. This is obsoleted by send_message.
 */
extern Success
broadcast (const String message)
{
    return send_message(0, message);
}

/*
 * Send a message
 */
extern Success
send_message (Conf_no recipient,
	      const String message)
{
    Conference *conf_c;
    unsigned short end;
    unsigned short ix;
    Success retval;

    CHK_LOGIN(FAILURE);

    if (s_strlen(message) > param.broadcast_len)
    {
	kom_errno = KOM_LONG_STR;
	return FAILURE;
    }

    if (recipient != 0 && cached_conf_exists(recipient) == FALSE)
    {
	kom_errno = KOM_UNDEF_CONF;
	return FAILURE;
    }
    
    if (recipient == 0 || cached_get_conf_type(recipient).letter_box == 1)
	return async_send_message(recipient, ACTPERS, message);
    else
    {
	/* The recipient is a conference.  Send the message to all
	   members of that conference. */

	GET_C_STAT(conf_c, recipient, FAILURE);
	end = conf_c->members.no_of_members;
	retval = FAILURE;
	for (ix = 0; ix < end; ix++)
	{
	    if (async_send_group_message(conf_c->members.members[ix].member,
					 recipient, ACTPERS, message) == OK) 
	    {
		retval = OK;
	    }
	}
	return retval;
    }
}


/*
 * Make LysKOM sync its files.
 */
extern Success
sync_kom (void) 
{
    cache_sync_all();
    dump_statistics();
    return OK;
}

/*
 * Close LysKOM. exit_val is (currently) not used. The database is synced.
 */
extern Success
shutdown_kom (int exit_val) 
{
    char *name;
    char *user;
    char *host;

    CHK_LOGIN(FAILURE);
    if ( !ENA(admin, 1) )
    {
	kom_errno = KOM_PERM;
	return FAILURE;
    }

    name = s_crea_c_str (active_connection->username);
    user = s_crea_c_str (active_connection->ident_user);
    host = s_crea_c_str (active_connection->hostname);
    log ("shutdown initiated by person %d (%s) via %s@%s.\n", 
	 ACTPERS, name, user, host);
    string_free(host);
    string_free(user);
    string_free(name);
    
    go_and_die = TRUE;
    return OK;
}
