/*
 * $Id: connections.h,v 0.18 1995/01/01 20:17:20 ceder Exp $
 * Copyright (C) 1991, 1992, 1993, 1994, 1995  Lysator Academic Computer Association.
 *
 * This file is part of the LysKOM server.
 * 
 * LysKOM is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 1, or (at your option) 
 * any later version.
 * 
 * LysKOM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with LysKOM; see the file COPYING.  If not, write to
 * Lysator, c/o ISY, Linkoping University, S-581 83 Linkoping, SWEDEN,
 * or the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, 
 * MA 02139, USA.
 *
 * Please mail bug reports to bug-lyskom@lysator.liu.se. 
 */
/*
 * $Id: connections.h,v 0.18 1995/01/01 20:17:20 ceder Exp $
 *
  * connections.h -- The top level of the communication packet.
  *
  * Requires kom-types.h, com.h
  */


extern volatile Bool go_and_die; /* Set this to TRUE to kill the server. */
extern volatile Bool do_sync_db; /* Set this to TRUE to force a sync via
				    signal. */

#define CONN_MAGIC_ALLOC	0x18F4AE74
#define CONN_MAGIC_FREE         0x67A7B144

typedef struct connection {
    unsigned long       magic;           /* Hey, this is like magic! */
    
    struct connection * prev;
    struct connection * next;
    
    /* The following are used by services.c (which has been split
       into many files such as conference.c and person.c). */

    Pers_no		pers_no;         /* 0 if not yet logged in. */
    Person	      * person;	         /* NULL if not yet logged in. */
    time_t		session_start;   /* the time when the current session
					    was started */
    Conf_no		cwc;	         /* Current working conference */
    String		what_am_i_doing;
    unsigned char	ena_level;       /* Enable level */
    String		username;        /* Given userid. */
    String		hostname;        /* Real hostname  */
    String		ident_user; 	 /* Real username */
    String              client_name;     /* See doc/clients.assigned */
    String              client_version;
    Bool		invisible; 	 /* Is this session invisible? */
    Bool		username_valid;	 /* Set once username is set. */

    /* The following are used by server/connections.c */
    
    struct mux         *mux;
    unsigned char		protocol;

    int			parse_pos;
    int			fnc_parse_pos;
    int			array_parse_pos;
    int			struct_parse_pos;
    int			string_parse_pos;

    int			ref_no;
    Call_header		function; /* Function to call. */

    /* Gather data in these variables. */
    
    long		num0;
    long		num1;
    long		num2;
    long		num3;
    String		c_string0; /* These strings are used for strings
				    * that are *NOT* freed by services.c.
				    * They are freed by free_parsed() in
				    * connections.c */
    String		c_string1;
    String		string0; /* This string is used for strings that
				   * are freed by the routines in services.c */
    Misc_info	       *c_misc_info_p;	/* Freead by free_parsed(). */
    Local_text_no      *c_local_text_no_p; /* Freead by free_parsed(). */
    Priv_bits		priv_bits;
    Conf_type		conf_type;
    struct tm		time;

    /* Protocol independent things. */
    
    String              unparsed;
    String_size		first_to_parse; /* Index into unparsed. */
    Bool		more_to_parse;	/* Any chance that there is anything
					   left in unparsed? */
    time_t		last_request;   /* Set every time a packet is
					    received from the client. */
    
    Session_no		session_no;      /* A unique number. */
} Connection;
/*
 * It is guaranteed that in the Connection struct the pers_no and person 
 * fields are either both 0/NULL, or both non-0/non-NULL.
 */


/*
 * Data for the isc-server.
 */
extern struct isc_mcb *kom_server_mcb;

/*
 * active_connection points to the info about the client who initiated this
 * call. This is set by connectins.c whenever a complete message is parsed
 * and a function in services.c is called.
 */
extern Connection * active_connection;


/*
 * This enum describes the result of a function in services.c.
 */
typedef enum {
    rt_number,			/* E.g. Pers_no, Conf_no, Text_no.
				   This is somewhat special in that the
				   function indicates an error by returning
				   0 and not FAILURE. */
    rt_success,			/* Only return a Success. */
    /* Functions of the following types returns a Success _and_ data
       of the indicated type. The corresponding types are found in
       kom-types.h */
    rt_person,
    rt_membership,
    rt_conf_list,
    rt_conf_no_list,
    rt_conference,
    rt_string,
    rt_mark_list,
    rt_text_stat,
    rt_text_list,
    rt_who_info_list,
    rt_who_info_list_old,
    rt_info,
    rt_membership_list,
    rt_member_list,
    rt_time_date,
    rt_session_info,
    rt_session_no,
    rt_text_no,
    rt_session_info_ident,
    rt_who_info_ident_list,
    rt_conf_z_info_list,
    rt_version_info
} Res_type;

/*
 * The result from a call is stored in this union.
 *
 * The pointers in these structures are normally set to point into
 * data from the cache, and should thus not be free()d. In those cases
 * where this actually contains newly allocated memory (and it is not
 * allocated via tmp_alloc()) it should be free()d in prot_a_reply() in
 * prot-a.c.
 */
typedef union {
        unsigned long		  number;
    	Person			  person;
	Membership		  membership;
	Conf_list_old		  conf_list;
	Conf_no_list		  conf_no_list;
	Conference		  conference;
	Mark_list		  mark_list;
	String			  string;
	Text_stat		  text_stat;
	Text_list		  text_list;
	Info			  info;
	Who_info_list		  who_info_list;
	Who_info_list_old	  who_info_list_old;
	Membership_list		  membership_list;
	Member_list		  member_list;
	time_t			  time_date;
	Session_info		  session_info;
	Session_no		  session_no;
	Text_no			  text_no;
	Who_info_ident_list	  who_info_ident_list;
	Session_info_ident	  session_info_ident;
	Conf_z_info_list          conf_z_info_list;
        Version_info              version_info;
} Result_holder;


typedef struct {
    Res_type  result;
    /* The function that is used to parse args for this function: */
    void      (*parser)(Connection *client);
} Fnc_descriptor;

extern const Fnc_descriptor fnc_defs[];

/*
 * This array holds number of calls to each service. It is dumped to the
 * log file every time the server syncs.
 */
extern unsigned long service_statistics[];

extern jmp_buf parse_env;

extern void
toploop(void);

extern void
dump_statistics(void);

extern void
logout_all_clients(void);

extern void
add_to_kill_list(Connection *conn);

