/*
 * $Id: mux.h,v 0.8 1995/01/01 20:17:23 ceder Exp $
 * Copyright (C) 1991, 1993, 1994, 1995  Lysator Academic Computer Association.
 *
 * This file is part of the LysKOM server.
 * 
 * LysKOM is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 1, or (at your option) 
 * any later version.
 * 
 * LysKOM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with LysKOM; see the file COPYING.  If not, write to
 * Lysator, c/o ISY, Linkoping University, S-581 83 Linkoping, SWEDEN,
 * or the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, 
 * MA 02139, USA.
 *
 * Please mail bug reports to bug-lyskom@lysator.liu.se. 
 */
/*
 * $Id: mux.h,v 0.8 1995/01/01 20:17:23 ceder Exp $
 *
** mux.h
*/

#ifndef __LYSKOM__MUX_H__
#define __LYSKOM__MUX_H__

#define MUX_OUTMSGSIZE	   1024

#define MUX_MAGIC_ALLOC    0x56AE17DB
#define MUX_MAGIC_FREE     0xF24A6BE7

typedef enum
{
  MUX_TYPE_CLIENT,
  MUX_TYPE_MUX
} Mux_type;



typedef struct
{
  int             id;     /* MUX client id */
  Connection    * conn;   /* LysKOM virtual Connection */

  struct isc_msg * outmsg;
} Mux_client;

typedef struct
{
    int			parse_pos;
    int			parse_pos_2;
    int			string_parse_pos;
    int			function; /* Function to call. */

    /* Gather data in these variables. */
    
    long		num;
    String		string;
    String              unparsed;
    String_size		first_to_parse; /* Index into unparsed. */
    Bool		more_to_parse;	/* Any chance that there is anything
					   left in unparsed? */
} Mux_parse;

typedef struct mux
{
  unsigned long     magic;    /* Magic number :-) */
  Mux_type          type;
  struct isc_scb  * scb;

  Mux_parse         parse;

  int               client_c;
  Mux_client      * client_v;
} Mux;


/*
** MUX support functions
*/

extern Mux *
mux_create(Mux_type type, IscSession *scb);
  
extern void
mux_destruct(Mux *mp);

extern int
mux_clients(Mux *mp);

extern void
mux_addclient(Mux *mp, int id, Connection *cp);

extern void
mux_delclient(Mux *mp, int id);

extern Mux_client *
mux_getclientbyid(Mux *mp, int id);

extern Mux_client *
mux_getclientbyconn(Mux *mp, Connection *cp);



/*
** MUX output functions
*/

extern int
mux_write(Connection  	       * cp,
	  const unsigned char  * buf,
	  size_t        	 size);

#ifdef HAVE_STDARG_H
extern int
mux_printf(Connection  * cp,
	   const char  * format,
	   ...);
#endif

extern int
mux_putc(int           chr,
	 Connection  * cp);

extern void
mux_flush(Connection  * cp);

extern void
mux_close(Connection  * cp);

extern void
mux_free_parsed(Mux *mp);

#endif
