/*
 * $Id: ramkomd.c,v 0.54 1994/12/31 15:16:19 ceder Exp $
 * Copyright (C) 1991, 1992, 1993, 1994  Lysator Academic Computer Association.
 *
 * This file is part of the LysKOM server.
 * 
 * LysKOM is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 1, or (at your option) 
 * any later version.
 * 
 * LysKOM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with LysKOM; see the file COPYING.  If not, write to
 * Lysator, c/o ISY, Linkoping University, S-581 83 Linkoping, SWEDEN,
 * or the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, 
 * MA 02139, USA.
 *
 * Please mail bug reports to bug-lyskom@lysator.liu.se. 
 */
/*
 * Detta {r serverns huvudprogram. Det kommer f|rhoppningsvis bli st|rre
 * {n det {r just nu...
 *
 * Created by Willf|r 31/3-90
 *
 * It has grown! /ceder
 */

static char *rcsid = "$Id: ramkomd.c,v 0.54 1994/12/31 15:16:19 ceder Exp $";
#include "rcs.h"
USE(rcsid);

#ifdef HAVE_LOCALE_H
#  include <locale.h>
#endif
#include <stdio.h>
#include <signal.h>
#ifdef HAVE_STDARG_H
#  include <stdarg.h>
#endif
#ifdef HAVE_STDLIB_H
#  include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#  include <string.h>
#endif
#include <sys/wait.h>
#include <time.h>
#ifdef HAVE_SYS_RESOURCE_H
#  include <sys/time.h>
#  include <sys/resource.h>
#endif
#include <unistd.h>
#include <setjmp.h>
#if defined(HAVE_SYS_PARAM_H) && !defined(HAVE_GETCWD)
# include <sys/param.h>
#endif
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <fcntl.h>

#include "exp.h"
#include "s-string.h"
#include "misc-types.h"
#include "kom-types.h"
#include "com.h"
#include "connections.h"
#include "internal-connections.h"
#include "kom-errno.h"
#include "isc-malloc.h"
#include "isc-interface.h"
#include "config.h"
#include "cache.h"
#include "string-malloc.h"
#include "lyskomd.h"
#include "log.h"
#include "internal-services.h"
#include "server/smalloc.h"
#include "kom-memory.h"
#include "conf-file.h"
#include "param.h"
#include "server-config.h"
#include "manipulate.h"
#include "version-info.h"

#if defined(HAVE_SETRLIMIT) && defined(RLIMIT_OFILE) && !defined(RLIMIT_NOFILE)
#  define RLIMIT_NOFILE RLIMIT_OFILE
#endif

struct kom_par param;
 
Kom_err		  kom_errno = KOM_NO_ERROR;
unsigned long	  err_stat = 0;


#ifndef NDEBUG
int	buglevel = 0;
#endif

EXPORT char datafilename[BUFSIZ];      /* Full pathname to the database file */
EXPORT char backupfilename[BUFSIZ];    /* Full pathname to the backup file */
EXPORT char textfilename[BUFSIZ];      /* Full pathname to the text file. */
EXPORT char statisticfile[BUFSIZ];     /* Dito to statistics file. */
EXPORT char pidfile[BUFSIZ];	       /* Dito pid-file. */
INTERNAL char memusefile[BUFSIZ];      /* Memory usage file. */

static IscSession *listen_client = NULL;      /* ISC listen identifier */
static IscSession *listen_mux    = NULL;      /*          -"-          */

static void dump_exit_statistics(void);

static void
server_init (char * client_port, char * mux_port)
{
    IscConfig config;
    IscAddress *isc_adr = NULL;
    
    /*
    ** Setup some parameters here
    */
    config.version = 1005;
    config.master.version = 1001;
    config.master.memfn.alloc = &isc_malloc_wrapper;
    config.master.memfn.realloc = &isc_realloc_wrapper;
    config.master.memfn.free = &isc_free_wrapper;
    config.master.abortfn = NULL; /* Use default abort function. */
    config.session.version = 1001;
    config.session.max.msgsize = -1; /* Use default sizes. */
    config.session.max.queuedsize = -1;
    config.session.max.dequeuelen = -1;
    config.session.max.openretries = -1;
    config.session.max.backlog = -1;

    kom_server_mcb  = isc_initialize(&config);
    if ( kom_server_mcb == NULL )
	restart_kom("server_init: can't isc_initialize()\n");

    listen_client = isc_listentcp(kom_server_mcb, NULL, client_port);
    if (listen_client == NULL)
        restart_kom("server_init: can't isc_listentcp(CLIENT)\n");

    isc_adr = isc_getladdress (listen_client);
    if (isc_adr == NULL)
	restart_kom("server_init(): can't isc_getladdress (listen_client)\n");

    param.num_ip_client_port = isc_getportnum (isc_adr);
    isc_freeaddress (isc_adr);
    
    listen_mux = isc_listentcp (kom_server_mcb, NULL, mux_port);
    if (listen_mux == NULL)
        restart_kom("server_init: can't isc_listentcp(MUX)\n");

    isc_adr = isc_getladdress (listen_mux);
    if (isc_adr == NULL)
	restart_kom("server_init(): can't isc_getladdress (listen_mux)\n");
    
    param.num_ip_mux_port = isc_getportnum (isc_adr);
    isc_freeaddress (isc_adr);

    log("Listening for clients on %d and muxes on %d.\n",
	param.num_ip_client_port, param.num_ip_mux_port);

    /*
     * Ignore SIGPIPE, which the server gets if it tries to write to a
     * socket and the client has died. The server will anyhow handle
     * this situation correct.
     */
    signal(SIGPIPE, SIG_IGN);
}

static void
init_data_base(const char *dir_base)
{
    sprintf(datafilename,   "%s/%s", dir_base, param.datafile_name);
    sprintf(backupfilename, "%s/%s", dir_base, param.backupfile_name);
    sprintf(textfilename, "%s/%s", dir_base, param.textfile_name);
    sprintf(statisticfile, "%s/%s", dir_base, param.statistic_name);
    sprintf(pidfile, "%s/%s", dir_base, param.pid_name);
    sprintf(memusefile, "%s/%s", dir_base, param.memuse_name);

    log("Database = %s\n",  datafilename);
    log("Backup = %s\n", backupfilename);
    
    if ( init_cache() == FAILURE )
	restart_kom ("Cannot find database.\n");
}

static void
sighandler_hup (int sig)
{
    log ("Signal HUP received. Shutting down server.\n");
    go_and_die = TRUE;
}

static void
sighandler_quit (int sig)
{
    log ("Signal QUIT received - syncing...\n");
    cache_sync_all();
    log ("Dumping core now.\n");
    abort();
}

static void
sighandler_usr1 (int sig)
{
    signal(SIGUSR1, sighandler_usr1);
    do_sync_db = TRUE;
}

static void
sighandler_usr2 (int sig)
{
    int		  child;

    log ("Signal USR2 received - will dump core now.  (Check that child dies.)\n");
    if ((child = fork()) == 0)
    {
	abort();
	log ("Abort() failed!!!");
	exit(1);
    }
    else if (child < 0)
    {
	log ("Couldn't fork.\n");
    }
    else
    {
	wait (NULL);
    }
}

static void
save_pid(void)
{
    FILE *fp;

    if ( (fp = fopen(pidfile, "w")) == NULL )
	return;

    fprintf(fp, "%ld\n", (long)getpid());
    fclose(fp);
}

static void
go_daemon(const char *prefix, const char *logf)
{
    pid_t child;
    int fd;
    char *logfile;

    if (getppid() != 1)
    {
	/* We were not invoked from /etc/inittab, so
	   disassociate from controlling terminal. */

#ifdef SIGTTOU
	signal(SIGTTOU, SIG_IGN);
#endif
#ifdef SIGTTIN
	signal(SIGTTIN, SIG_IGN);
#endif
#ifdef SIGTSTP
	signal(SIGTSTP, SIG_IGN);
#endif

	child = fork();
	if (child < 0)
	    restart_kom("fork failed: %d\n", errno);
	else if (child > 0)
	    exit (0);		/* parent */
	
	setsid();
    }

    logfile = smalloc(strlen(prefix) + 1 + strlen(logf) + 1);
    sprintf(logfile, "%s/%s", prefix, logf);
    /* Close all file descriptors */
    for (fd = 0; fd < MAX_NO_OF_CONNECTIONS + PROTECTED_FDS; fd++)
	close(fd);
    if (open("/dev/null", O_RDONLY) != 0
	||open("/dev/null", O_WRONLY) != 1
	||open(logfile, O_WRONLY|O_CREAT|O_APPEND, 0644) != 2)
    {
	/* Kinda stupid to try to log an error message now, but
	   who knows? */
	restart_kom("open of log file failed: %d\n", errno);
    }
    sfree(logfile);
    log("*** Version %s (process %d) coming up.\n",
	kom_version_info.server_version, getpid());
}

void
initialize(const char *config_file)
{
#if defined(HAVE_SETRLIMIT) && defined(RLIMIT_NOFILE)
    struct rlimit rlim;
#endif

    read_configuration(config_file);

#ifdef HAVE_LOCALE_H
    if (param.use_locale != NULL)
	if (setlocale(LC_CTYPE, param.use_locale) == NULL)
	{
	    fprintf(stderr, "setlocale: ");
	    perror(param.use_locale);
	    exit(1);
	}
#else
    if (param.use_locale != NULL)
    {
	fprintf(stderr, "locale not supported in your environment.\n");
	exit(1);
    }
#endif

#if defined(HAVE_SETRLIMIT) && defined(RLIMIT_NOFILE)
    if (param.no_files != -1)
    {
	if (getrlimit(RLIMIT_NOFILE, &rlim) < 0)
	{
	    perror("getrlimit(RLIMIT_NOFILE) failed");
	    exit(1);
	}
	if (param.no_files > rlim.rlim_max)
	{
	    fprintf(stderr, "attempt to raise open files from %d to %d, "
		    "but only %d is allowed\n", rlim.rlim_cur, param.no_files,
		    rlim.rlim_max);
	    rlim.rlim_cur = rlim.rlim_max;
	}
	else
	    rlim.rlim_cur = param.no_files;
	
	if (setrlimit(RLIMIT_NOFILE, &rlim) < 0)
	{
	    perror("setrlimit failed");
	    exit(1);
	}
    }
#else
    if (param.no_files != -1)
    {
	fprintf(stderr,
		"can't alter number of open files in your environment.\n");
	exit(1);
    }
#endif

    /* Find out how many connections we can handle. */
#if defined (HAVE_SYSCONF)
    MAX_NO_OF_CONNECTIONS = sysconf(_SC_OPEN_MAX) - PROTECTED_FDS;
#else
# ifdef HAVE_GETDTABLESIZE
    MAX_NO_OF_CONNECTIONS = getdtablesize() - PROTECTED_FDS;
# else
    /* If we don't have getdtablesize or sysconf, MAX_NO_OF_CONNECTIONS is
       set at compile time. */
# endif
#endif

    go_daemon(param.dbase_dir, param.logfile_name);
    server_init(param.ip_client_port, param.ip_mux_port);
    init_data_base(param.dbase_dir);
}

/* Stop complaint from gcc 2.0 about "no previous prototype for `main'". */
int main(int argc, char **argv);

	
int
main (int    argc,
      char **argv)
{
    int i;
    char *default_config_file;
    char *config_file;

    log("*** Version %s (process %d) started.\n",
	kom_version_info.server_version, getpid());

#ifdef ENCRYPT_PASSWORDS
    /* Seed the random number generator. */
    srand(time(NULL) + getpid());
#endif

    /* Initialize the string handling package. */
    s_set_storage_management(string_malloc, string_realloc, string_free);

    /* Parse command line arguments. */
    for (i = 1; i < argc && argv[i][0] == '-'; i++)
	switch (argv[i][1])
	{
	case 'd':
	    buglevel++;
	    break;

	default:
	    restart_kom("usage: %s [-d ...] [config-file]\n", argv[0]);
	}


    /* Read in the configuration file. */

    default_config_file = smalloc(strlen(DEFAULT_DBASE_DIR) +
				  strlen(CONFIG_FILE) + 2);
    sprintf(default_config_file, "%s/%s", DEFAULT_DBASE_DIR, CONFIG_FILE);
    if (i < argc)
	config_file = argv[i++];
    else
	config_file = default_config_file;

    if (i < argc)
	    restart_kom("usage: %s [-d ...] [config-file]\n", argv[0]);

    signal(SIGHUP, sighandler_hup);
    signal(SIGQUIT, sighandler_quit);
    signal(SIGUSR1, sighandler_usr1);
    signal(SIGUSR2, sighandler_usr2);
  
    initialize(config_file);	/* Read config, listen, and start db */
    sfree(default_config_file);

    save_pid();

    toploop();

    logout_all_clients();
    isc_shutdown(kom_server_mcb);
    cache_sync_all();

    dump_exit_statistics();

    log("%s terminated normally.\n", argv[0]);
    
    return 0;
}

static void 
dump_exit_statistics(void)
{
    FILE *stat_file;
    time_t now;

    time(&now);
    stat_file = fopen(memusefile, "a");

    if ( stat_file == NULL )
	restart_kom("Can't open file to save memory usage to.\n");

    fprintf(stat_file, "\nLysKOM Server going down at %s\n", ctime(&now));

    dump_cache_stats (stat_file);

    free_all_tmp();
    free_all_cache();
    free_all_jubel();

    dump_smalloc_counts(stat_file);
    dump_alloc_counts(stat_file);
    dump_cache_mem_usage(stat_file);
    dump_string_alloc_counts(stat_file);
    dump_allocated_connections(stat_file);
    dump_isc_alloc_counts(stat_file);
    fclose (stat_file);
}
    
void
#if defined(HAVE_VFPRINTF) && defined(HAVE_STDARG_H)
restart_kom(const char * format, ...)
#else
restart_kom(format,
	    a, b, c, d, e, f, g)
     const char *format;
     int a, b, c, d, e, f, g;
#endif
{
#if defined(HAVE_VFPRINTF) && defined(HAVE_STDARG_H)
    va_list AP;
#endif
#if defined(HAVE_GETCWD)
    char pathname[1026];
#else /* !defined(HAVE_GETCWD) */
    char *getwd(char *); /* +++ should be in sys/param.h */
# if defined(MAXPATHLEN)
    char pathname[MAXPATHLEN];
# else
    char pathname[1024];
# endif /* !defined(MAXPATHLEN) */
#endif /* !defined(HAVE_GETCWD) */

#if defined(HAVE_VFPRINTF) && defined(HAVE_STDARG_H)
    va_start(AP, format);
    logv(format, AP);
    va_end(AP);
#else
    log(format, a, b, c, d, e, f, g);
#endif

    log("Previous message is fatal. Will dump core now.\n");

#ifdef HAVE_GETCWD
    /* getcwd is POSIX, so try that first. */
    if (getcwd(pathname, 1026) == NULL)
	log("getcwd failed: errno %d\n", errno);
    else
	log("Search for the core in %s\n", pathname);
#else
    if ( getwd(pathname) == NULL )
	log("getwd failed: %s\n", pathname);
    else
	log("Search for the core in %s\n", pathname);
#endif
    
    abort();
}
