/*
 * $Id: updateLysKOM.c,v 1.4 1994/04/05 20:30:00 ceder Exp $
 * Copyright (C) 1994  Lysator Academic Computer Association.
 *
 * This file is part of the LysKOM server.
 * 
 * LysKOM is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 1, or (at your option) 
 * any later version.
 * 
 * LysKOM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with LysKOM; see the file COPYING.  If not, write to
 * Lysator, c/o ISY, Linkoping University, S-581 83 Linkoping, SWEDEN,
 * or the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, 
 * MA 02139, USA.
 *
 * Please mail bug reports to bug-lyskom@lysator.liu.se. 
 */

#include <stdio.h>
#ifdef HAVE_STDLIB_H
#  include <stdlib.h>
#endif
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#ifdef HAVE_STRING_H
#  include <string.h>
#else
#  ifdef HAVE_STRINGS_H
#    include <strings.h>
#  endif
#endif
#ifndef HAVE_STRCHR
#  define strchr index
#endif
#include <time.h>
#include <unistd.h>

#include "ldifftime.h"

void
checkstatus(FILE *fp,
	    long pid,
	    char *progname)
{
    char lbuf[80];
    char buf2[300];
    struct stat sbuf;

    if (fstat(fileno(fp), &sbuf) < 0)
    {
	perror("updateLysKOM: stat failed");
	exit(1);
    }

    if (pid != 0)
    {
	if (kill(pid, SIGHUP) != 0)
	{
	    if (errno != ESRCH)
	    {
		fprintf(stderr, "%s: kill(%ld, SIGHUP) failed", progname, pid);
		perror("");
		exit(1);
	    }
	}
	else
	{
	    /* The signal got through... */
	    /* cron 10 minutes */
	    if (ldifftime(time(NULL), sbuf.st_mtime) > 1260)
	    {
		/* ...more than once.  Strange. */
		fprintf(stderr, "%s: Told lyskomd to shut down.  Twice.", 
			progname);
		exit(1);
	    }
	}
    }

    if (ldifftime(time(NULL), sbuf.st_mtime) < 3600)
    {
	/* NOP */
    }
    else if (ldifftime(time(NULL), sbuf.st_mtime) < 2*3600)
    {
	/* Send mail to creator of file, the first line in the file. */

	if (fgets(lbuf, 80, fp) == NULL)
	{
	    fprintf(stderr, 
		    "updateLysKOM: LysKOM has been down for a short while\n");
	    exit(2);
	}
	if (strchr(lbuf, '\n'))
	    *strchr(lbuf, '\n') = '\0';
	sprintf(buf2, "echo 'Reminder: LysKOM is still not running.'|Mail %s",
		lbuf);
	if (system(buf2) != 0)
	{
	    perror("updateLysKOM: system failed");
	    exit(1);
	}
    }
    else
    {
	fprintf(stderr,
		"updateLysKOM: LysKOM has been down for a long time\n");
	exit(2);
    }
}
	

int
main (int    argc,
      char **argv)
{
    int i;
    FILE *fp;
    long pid;

    if (getuid() == 0 || geteuid() == 0)
    {
	fprintf(stderr, "%s: this program should run as lyskom, not root\n",
		argv[0]);
	exit(1);
    }

    /* Parse command line arguments. */
    for (i = 1; i < argc && argv[i][0] == '-'; i++)
	switch (argv[i][1])
	{
	case 'V':
	case 'v':
	    fprintf(stderr, "updateLysKOM %d $Revision: 1.4 $\n",
#include "version.incl"
		    );
	    exit(0);

	default:
	    fprintf(stderr, "usage: %s [-V]\n", argv[0]);
	}

    fp = fopen(DEFAULT_PREFIX "/etc/pid", "r");
    if (fp == NULL)
    {
	fprintf(stderr, "%s: warning: %s: ", argv[0], 
		DEFAULT_PREFIX "/etc/pid");
	perror("");
	pid = 0;
    }
    else
    {
	pid = 0;
	fscanf(fp, "%lu", &pid);
	if (pid < 2)
	{
	    fprintf(stderr, "%s: got pid %ld.\n", argv[0], pid);
	    exit(1);
	}
    }
	
    fp = fopen(DEFAULT_PREFIX "/etc/status", "r");
    if (fp != NULL)
    {
	checkstatus(fp, pid, argv[0]);
    }
    else
    {
	errno = ESRCH;		/* Set sane default if pid==0. */
	if (pid == 0 || kill(pid, SIGUSR1) != 0)
	{
	    if (errno != ESRCH)
	    {
		fprintf(stderr, "%s: kill(%ld, SIGUSR1) failed", argv[0], pid);
		perror("");
		exit(1);
	    }
	    execl(DEFAULT_PREFIX "/bin/lyskomd", "lyskomd", (char *)0);
	    fprintf(stderr, "%s: execl() failed: ", argv[0]);
	    perror("");
	    exit(1);
	}
    }

    exit(0);
}
